==> oscars_mgmnt_clms-0.1.1-py3-none-any.whl: oscars_mgmnt_clms/parser/filename.py <==
import os
from typing import Tuple, Optional

import ciso8601
import shapely

from oscars_mgmnt_clms.file_object import ObjectInfo, FileInfo
from oscars_mgmnt_clms.parser.base import BaseParser
import re
from datetime import datetime as dt
import time
from abc import ABC, abstractmethod


class FilenameParser(BaseParser, ABC):
    def convert(self, product_file_info):
        values = {}
        # common properties for filename parsers
        fn = os.path.abspath(product_file_info.get_product_path())
        # filename without extension
        file_name = os.path.splitext(os.path.basename(fn))[0]
        file_name_parts = file_name.split("_")

        nominaldatetime = parse_fname_date(file_name_parts[3])
        values["nominaldatetime"] = format_date(nominaldatetime)

        # collection identifier is constructed from the collection-folder
        path_elem = os.path.dirname(fn).split(os.path.sep)

        # assuming files are stored ...
        nominal_year_str = str(nominaldatetime.year)
        #  ...  either in 3rd level below collection, ie collection/yyyy/yyyymmdd/file
        if path_elem[-2] == nominal_year_str and path_elem[-1].startswith(
            nominal_year_str
        ):
            collection_dir = path_elem[-3]
        else:
            #  ...  or in lowest folder level (no organization per date, e.g. static masks)
            collection_dir = path_elem[-1]

        parent_identifier = "_".join(["clms_global", collection_dir, self.filetype])
        values["parentidentifier"] = parent_identifier

        values["creationtime"] = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        values["catingestiondate"] = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        values["updatetime"] = time.strftime(
            "%Y-%m-%dT%H:%M:%SZ", product_file_info.storage.get_time_modified(fn)
        )

        # specific treatment
        specific_infos = self.splitted_treatment(product_file_info, values)

        # Generation of the output
        results = {}

        if self.jinja_template is None:
            self.with_template(self.template)

        for specific in specific_infos:
            result = self.jinja_template.render(
                dict=specific.values,
                dirpath=specific.product_file_info.get_product_path().rpartition("/")[
                    0
                ],
                products=specific.product_file_info.datas,
                previews=specific.product_file_info.previews,
                metadatas=specific.product_file_info.metadatas,
                related=specific.product_file_info.related,
                platforms=specific.values["platforms"],
            )
            results[specific.uid] = result
        return results

    @property
    @abstractmethod
    def filetype(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def mimetype(self) -> str:
        raise NotImplementedError

    def fix_products(self, fileobject: ObjectInfo):
        for info in list(fileobject.related):
            # move files from related to data section
            if os.path.splitext(info.name)[1] in [".tiff", ".nc", ".json"]:
                fileobject.related.remove(info)
                fileobject.datas.append(info)

        for info in fileobject.datas:
            info.title = self.get_file_title(info)
            info.type = self.mimetype

    def get_file_title(self, info: FileInfo) -> str:
        title = os.path.basename(info.name).split("_")[2].split("-")[0]
        re_match = re.search(r"\d", title)
        if re_match:
            # if needed, use all characters up to first digit (eg NDVI300 -> NDVI)
            title = title[: re_match.start()]
        return title


def parse_fname_date(fname_date: str) -> dt:
    """
    Parse the date from the filename of a product.

    Nominal dates in filenames can take one of three forms:
        - YYYYMMDDhhmm (12 digits): the most common case
        - YEAR1----YEAR2 (4 digits, 4 hyphens, 4 digits) -> nominal date = Jan 1 of YEAR1:
            e.g. SWI static
        - YEAR1-YEAR2-MMDD (4 digits, hyphen, 4 digits, hyphen, 4 digits) -> nominal date = YEAR1/MM/DD:
            e.g. NDVI statistics
    """
    if re.search("[0-9]{12}", fname_date):
        return dt.strptime(fname_date, "%Y%m%d%H%M")
    elif re.search("[0-9]{4}----[0-9]{4}", fname_date):
        return dt.strptime(fname_date[0:4], "%Y")
    else:
        return dt.strptime(fname_date[0:4] + fname_date[-4:], "%Y%m%d")


def get_geometry(bbox: Tuple[float, float, float, float]) -> str:
    geometry = shapely.box(*bbox)
    geometry = shapely.make_valid(geometry)
    return shapely.to_geojson(geometry)


def parse_date(date_str: str) -> dt:
    """
    Parse a date string into a datetime object. Multiple date formats are supported.
    First, the ISO8601 format will be tried. If this does not match, other custom formats are tried.
    Raises a ValueError if the date string cannot be parsed.

    :param date_str: the date string to parse
    """
    try:
        return ciso8601.parse_datetime(date_str)
    except ValueError:
        pass
    for fmt in ("%d-%b-%Y %H:%M:%S", "%Y%m%dT%H:%M:%SZ"):
        try:
            return dt.strptime(date_str, fmt)
        except ValueError:
            pass
    raise ValueError(f"No valid date format found for '{date_str}'")


def format_date(date: dt) -> str:
    """
    Format a datetime object as a zulu time string.
    :param date: the datetime object to format
    :return: the formatted datetime string
    """
    timespec = "auto"
    if date.microsecond:
        # prefer milliseconds if possible
        timespec = "milliseconds" if date.microsecond % 1000 == 0 else "microseconds"
    if date.tzinfo is None:
        return date.isoformat(timespec=timespec) + "Z"
    else:
        return date.isoformat(timespec=timespec).replace("+00:00", "Z")


def decode_platforms(platformstr):
    # decodes the netCDF :platform attribute into a list of one or more
    # platforms, with consistent formatting

    # examples:
    # Proba-V => [PROBA-V]
    # SPOT 4 => [SPOT-4]
    # METOP A+B => [METOP-A, METOP-B]
    # Sentinel-1 A+B => [SENTINEL-1A, SENTINEL-1B]
    # METOP A+B, Sentinel-1 A+B => [METOP-A, METOP-B, SENTINEL-1A, SENTINEL-1B]
    # SENTINEL-3A AND SENTINEL-3B => [SENTINEL-3A, SENTINEL-3B]
    # SENTINEL3A AND SENTINEL3B => [SENTINEL-3A, SENTINEL-3B]
    # SENTINEL 3 (A AND B) => [SENTINEL-3A, SENTINEL-3B]
    # TERRA (EOS AM-1) => [TERRA]
    # N/A => [N/A]

    platformstr = (
        platformstr.upper().replace("_", "-").replace("SUOMI-NPP", "SUOMI NPP").strip()
    )

    # support for platforms with LIE 250m with old name for TERRA/AQUA
    # TERRA (EOS AM-1) => [TERRA]
    for modis in ["AQUA", "TERRA"]:
        if modis in platformstr and " (" in platformstr:
            platformstr = platformstr.split(" (")[0]

    split_comma = platformstr.split(",")
    platforms = [x.strip() for x in split_comma]

    # Support for "AND" between platform names or platform serials
    # e.g. "SENTINEL-3A AND SENTINEL-3B" or "SENTINEL 3 (A AND B)"
    bracket_serials_regex = "\(([A-Z]) AND ([A-Z])\)"
    if " AND " in platformstr:
        repl_platforms = []
        for platf in platforms:
            regex_match_part = re.search(bracket_serials_regex, platf)
            if regex_match_part:
                (platformbase, platformserial) = platf.split(" (")
                platformbase = platformbase.replace(" ", "-")
                sep = "-"
                if "-" in platformbase:
                    sep = ""
                repl_platforms.extend(
                    [
                        platformbase + sep + serial
                        for serial in regex_match_part.groups()
                    ]
                )
            else:
                repl_platforms.extend(platf.split(" AND "))
        platforms = repl_platforms

    # Support for SSM/SWI multi-mission products with eg "METOP B+C"
    if "+" in platformstr:
        repl_platforms = []
        for platf in platforms:
            if "+" in platf:
                (platformbase, platformserial) = platf.split(" ")
                split_serial = platformserial.split("+")
                sep = "-"
                if "-" in platformbase:
                    sep = ""
                repl_platforms.extend(
                    [platformbase + sep + serial for serial in split_serial]
                )
            else:
                repl_platforms.extend(platf)
        platforms = repl_platforms

    # We covered the AND and + syntaxes, all remaining spaces are between platforms
    # with exception of "SUOMI NPP"
    repl_platforms = []
    for platf in platforms:
        if " " in platf and not "SUOMI NPP" == platf:
            repl_platforms.extend(platf.split(" "))
        else:
            repl_platforms.append(platf)
        platforms = repl_platforms

    # Support for LST, eg MSG4 -> MSG-4
    geos = ["GOES", "MSG", "HIMAWARI", "MTSAT"]
    lst_regex = "(MSG|MTSAT|HIMAWARI|GOES)([0-9]+|E|R)"
    if any([x in platformstr for x in geos]):
        repl_platforms = []
        for platf in platforms:
            regex_match = re.match(lst_regex, platf)
            if regex_match:
                repl_platforms.append("-".join(regex_match.groups()))
            else:
                repl_platforms.append(platf)
        platforms = repl_platforms

    # Hyphenate SENTINELs, eg SENTINEL3A => SENTINEL-3A
    sentinel_regex = "(SENTINEL)([0-6][A-Z]+)"
    if "SENTINEL" in platformstr:
        repl_platforms = []
        for platf in platforms:
            if "SENTINEL" in platf:
                regex_match = re.match(sentinel_regex, platf)
                if regex_match:
                    repl_platforms.append("-".join(regex_match.groups()))
                else:
                    repl_platforms.append(platf)
            else:
                repl_platforms.append(platf)
        platforms = repl_platforms

    return platforms


def half_pixel_shift(
    bbox: Tuple[float, float, float, float],
    x_size: int,
    y_size: int,
) -> Tuple[float, float, float, float]:
    """
    Apply a half-pixel shift on the bounding box coordinates.

    :param bbox: bounding box as tuple
    :param x_size: size of the raster in the x dimension
    :param y_size: size of the raster in the y dimension
    """
    west_center, south_center, east_center, north_center = bbox
    step = (east_center - west_center) / (x_size - 1)
    # improve final decimals for SWI and PROBA-V/Sentinel-3 vegetation products
    for realstep in [0.1, 1.0 / 112, 1.0 / 336]:
        precision = 10
        if round(realstep, precision) == round(step, precision):
            step = realstep
            east_center = west_center + ((x_size - 1) * step)
            south_center = north_center - ((y_size - 1) * step)
    halfpixel = step / 2.0

    # shift half pixel towards from center to edge of pixel
    west = west_center - halfpixel
    south = south_center - halfpixel
    east = east_center + halfpixel
    north = north_center + halfpixel

    return west, south, east, north


def fix_bbox(
    bbox: Tuple[float, float, float, float],
    rounding: int,
    x_size: Optional[int],
    y_size: Optional[int],
) -> Tuple[float, float, float, float]:
    """
    Fix the bounding box, by applying a half-pixel shift if necessary and rounding the coordinates to the specified precision.
    The x and y size can be None, then no half-pixel shift will be applied.

    :param bbox: bounding box
    :param rounding: rounding precision
    :param x_size: size of the raster in the x dimension
    :param y_size: size of the raster in the y dimension
    :return: fixed bounding box
    """
    if x_size is not None and y_size is not None:
        bbox = half_pixel_shift(bbox, x_size, y_size)

    bbox = tuple(map(lambda c: round(c, rounding), bbox))
    return bbox

==> oscars_mgmnt_clms-0.1.9.dev20240710+112-py3-none-any.whl: oscars_mgmnt_clms/parser/filename.py <==
import os
from typing import Tuple, Optional

import ciso8601
import shapely

from oscars_mgmnt_clms.file_object import ObjectInfo
from oscars_mgmnt_clms.parser.base import BaseParser
import re
from datetime import datetime as dt
import time
from abc import ABC, abstractmethod


class FilenameParser(BaseParser, ABC):
    def convert(self, product_file_info):
        values = {}
        # common properties for filename parsers
        fn = os.path.abspath(product_file_info.get_product_path())
        # filename without extension
        file_name = os.path.splitext(os.path.basename(fn))[0]

        nominaldatetime = self.get_nominal_date(file_name)
        values["nominaldatetime"] = format_date(nominaldatetime)

        # collection identifier is constructed from the collection-folder
        path_elem = os.path.dirname(fn).split(os.path.sep)

        # assuming files are stored ...
        nominal_year_str = str(nominaldatetime.year)
        #  ...  either in 3rd level below collection, ie collection/yyyy/yyyymmdd/file
        if path_elem[-2] == nominal_year_str and path_elem[-1].startswith(
            nominal_year_str
        ):
            collection_dir = path_elem[-3]
        else:
            #  ...  or in lowest folder level (no organization per date, e.g. static masks)
            collection_dir = path_elem[-1]

        parent_identifier = "_".join(["clms_global", collection_dir, self.filetype])
        values["parentidentifier"] = parent_identifier

        values["creationtime"] = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        values["catingestiondate"] = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        values["updatetime"] = time.strftime(
            "%Y-%m-%dT%H:%M:%SZ", product_file_info.storage.get_time_modified(fn)
        )

        # specific treatment
        specific_infos = self.splitted_treatment(product_file_info, values)

        # Generation of the output
        results = {}

        if self.jinja_template is None:
            self.with_template(self.template)

        for specific in specific_infos:
            result = self.jinja_template.render(
                dict=specific.values,
                dirpath=specific.product_file_info.get_product_path().rpartition("/")[
                    0
                ],
                products=specific.product_file_info.datas,
                previews=specific.product_file_info.previews,
                metadatas=specific.product_file_info.metadatas,
                related=specific.product_file_info.related,
                platforms=specific.values["platforms"],
            )
            results[specific.uid] = result
        return results

    def get_nominal_date(self, file_name: str):
        file_name_parts = file_name.split("_")
        nominaldatetime = parse_fname_date(file_name_parts[3])
        return nominaldatetime

    @property
    @abstractmethod
    def filetype(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def mimetype(self) -> str:
        raise NotImplementedError

    def fix_products(self, fileobject: ObjectInfo):
        for info in list(fileobject.related):
            # move files from related to data section
            if os.path.splitext(info.name)[1] in [".tiff", ".nc", ".json"]:
                fileobject.related.remove(info)
                fileobject.datas.append(info)

        for info in fileobject.datas:
            info.title = self.get_file_title(info.name)
            info.type = self.mimetype

    def get_file_title(self, filename: str) -> str:
        title = os.path.basename(filename).split("_")[2].split("-")[0]
        re_match = re.search(r"\d", title)
        if re_match:
            # if needed, use all characters up to first digit (eg NDVI300 -> NDVI)
            title = title[: re_match.start()]
        return title


def parse_fname_date(fname_date: str) -> dt:
    """
    Parse the date from the filename of a product.

    Nominal dates in filenames can take one of three forms:
        - YYYYMMDDhhmm (12 digits): the most common case
        - YEAR1----YEAR2 (4 digits, 4 hyphens, 4 digits) -> nominal date = Jan 1 of YEAR1:
            e.g. SWI static
        - YEAR1-YEAR2-MMDD (4 digits, hyphen, 4 digits, hyphen, 4 digits) -> nominal date = YEAR1/MM/DD:
            e.g. NDVI statistics
    """
    if re.search("[0-9]{12}", fname_date):
        return dt.strptime(fname_date, "%Y%m%d%H%M")
    elif re.search("[0-9]{4}----[0-9]{4}", fname_date):
        return dt.strptime(fname_date[0:4], "%Y")
    else:
        return dt.strptime(fname_date[0:4] + fname_date[-4:], "%Y%m%d")


def get_geometry(bbox: Tuple[float, float, float, float]) -> str:
    geometry = shapely.box(*bbox)
    geometry = shapely.make_valid(geometry)
    return shapely.to_geojson(geometry)


def parse_date(date_str: str) -> dt:
    """
    Parse a date string into a datetime object. Multiple date formats are supported.
    First, the ISO8601 format will be tried. If this does not match, other custom formats are tried.
    Raises a ValueError if the date string cannot be parsed.

    :param date_str: the date string to parse
    """
    try:
        return ciso8601.parse_datetime(date_str)
    except ValueError:
        pass
    for fmt in ("%d-%b-%Y %H:%M:%S", "%Y%m%dT%H:%M:%SZ"):
        try:
            return dt.strptime(date_str, fmt)
        except ValueError:
            pass
    raise ValueError(f"No valid date format found for '{date_str}'")


def format_date(date: dt) -> str:
    """
    Format a datetime object as a zulu time string.
    :param date: the datetime object to format
    :return: the formatted datetime string
    """
    timespec = "auto"
    if date.microsecond:
        # prefer milliseconds if possible
        timespec = "milliseconds" if date.microsecond % 1000 == 0 else "microseconds"
    if date.tzinfo is None:
        return date.isoformat(timespec=timespec) + "Z"
    else:
        return date.isoformat(timespec=timespec).replace("+00:00", "Z")


def decode_platforms(platformstr):
    # decodes the netCDF :platform attribute into a list of one or more
    # platforms, with consistent formatting

    # examples:
    # Proba-V => [PROBA-V]
    # SPOT 4 => [SPOT-4]
    # METOP A+B => [METOP-A, METOP-B]
    # Sentinel-1 A+B => [SENTINEL-1A, SENTINEL-1B]
    # METOP A+B, Sentinel-1 A+B => [METOP-A, METOP-B, SENTINEL-1A, SENTINEL-1B]
    # SENTINEL-3A AND SENTINEL-3B => [SENTINEL-3A, SENTINEL-3B]
    # SENTINEL3A AND SENTINEL3B => [SENTINEL-3A, SENTINEL-3B]
    # SENTINEL 3 (A AND B) => [SENTINEL-3A, SENTINEL-3B]
    # TERRA (EOS AM-1) => [TERRA]
    # N/A => [N/A]

    platformstr = (
        platformstr.upper().replace("_", "-").replace("SUOMI-NPP", "SUOMI NPP").strip()
    )

    # support for platforms with LIE 250m with old name for TERRA/AQUA
    # TERRA (EOS AM-1) => [TERRA]
    for modis in ["AQUA", "TERRA"]:
        if modis in platformstr and " (" in platformstr:
            platformstr = platformstr.split(" (")[0]

    split_comma = platformstr.split(",")
    platforms = [x.strip() for x in split_comma]

    # Support for "AND" between platform names or platform serials
    # e.g. "SENTINEL-3A AND SENTINEL-3B" or "SENTINEL 3 (A AND B)"
    bracket_serials_regex = "\(([A-Z]) AND ([A-Z])\)"
    if " AND " in platformstr:
        repl_platforms = []
        for platf in platforms:
            regex_match_part = re.search(bracket_serials_regex, platf)
            if regex_match_part:
                (platformbase, platformserial) = platf.split(" (")
                platformbase = platformbase.replace(" ", "-")
                sep = "-"
                if "-" in platformbase:
                    sep = ""
                repl_platforms.extend(
                    [
                        platformbase + sep + serial
                        for serial in regex_match_part.groups()
                    ]
                )
            else:
                repl_platforms.extend(platf.split(" AND "))
        platforms = repl_platforms

    # Support for SSM/SWI multi-mission products with eg "METOP B+C"
    if "+" in platformstr:
        repl_platforms = []
        for platf in platforms:
            if "+" in platf:
                (platformbase, platformserial) = platf.split(" ")
                split_serial = platformserial.split("+")
                sep = "-"
                if "-" in platformbase:
                    sep = ""
                repl_platforms.extend(
                    [platformbase + sep + serial for serial in split_serial]
                )
            else:
                repl_platforms.extend(platf)
        platforms = repl_platforms

    # We covered the AND and + syntaxes, all remaining spaces are between platforms
    # with exception of "SUOMI NPP"
    repl_platforms = []
    for platf in platforms:
        if " " in platf and not "SUOMI NPP" == platf:
            repl_platforms.extend(platf.split(" "))
        else:
            repl_platforms.append(platf)
        platforms = repl_platforms

    # Support for LST, eg MSG4 -> MSG-4
    geos = ["GOES", "MSG", "HIMAWARI", "MTSAT"]
    lst_regex = "(MSG|MTSAT|HIMAWARI|GOES)([0-9]+|E|R)"
    if any([x in platformstr for x in geos]):
        repl_platforms = []
        for platf in platforms:
            regex_match = re.match(lst_regex, platf)
            if regex_match:
                repl_platforms.append("-".join(regex_match.groups()))
            else:
                repl_platforms.append(platf)
        platforms = repl_platforms

    # Hyphenate SENTINELs, eg SENTINEL3A => SENTINEL-3A
    sentinel_regex = "(SENTINEL)([0-6][A-Z]+)"
    if "SENTINEL" in platformstr:
        repl_platforms = []
        for platf in platforms:
            if "SENTINEL" in platf:
                regex_match = re.match(sentinel_regex, platf)
                if regex_match:
                    repl_platforms.append("-".join(regex_match.groups()))
                else:
                    repl_platforms.append(platf)
            else:
                repl_platforms.append(platf)
        platforms = repl_platforms

    return platforms


def half_pixel_shift(
    bbox: Tuple[float, float, float, float],
    x_size: int,
    y_size: int,
) -> Tuple[float, float, float, float]:
    """
    Apply a half-pixel shift on the bounding box coordinates.

    :param bbox: bounding box as tuple
    :param x_size: size of the raster in the x dimension
    :param y_size: size of the raster in the y dimension
    """
    west_center, south_center, east_center, north_center = bbox
    step = (east_center - west_center) / (x_size - 1)
    # improve final decimals for SWI and PROBA-V/Sentinel-3 vegetation products
    for realstep in [0.1, 1.0 / 112, 1.0 / 336]:
        precision = 10
        if round(realstep, precision) == round(step, precision):
            step = realstep
            east_center = west_center + ((x_size - 1) * step)
            south_center = north_center - ((y_size - 1) * step)
    halfpixel = step / 2.0

    # shift half pixel towards from center to edge of pixel
    west = west_center - halfpixel
    south = south_center - halfpixel
    east = east_center + halfpixel
    north = north_center + halfpixel

    return west, south, east, north


def fix_bbox(
    bbox: Tuple[float, float, float, float],
    rounding: int,
    x_size: Optional[int],
    y_size: Optional[int],
) -> Tuple[float, float, float, float]:
    """
    Fix the bounding box, by applying a half-pixel shift if necessary and rounding the coordinates to the specified precision.
    The x and y size can be None, then no half-pixel shift will be applied.

    :param bbox: bounding box
    :param rounding: rounding precision
    :param x_size: size of the raster in the x dimension
    :param y_size: size of the raster in the y dimension
    :return: fixed bounding box
    """
    if x_size is not None and y_size is not None:
        bbox = half_pixel_shift(bbox, x_size, y_size)

    bbox = tuple(map(lambda c: round(c, rounding), bbox))
    return bbox
